Index: src/ia32/stub-cache-ia32.cc |
=================================================================== |
--- src/ia32/stub-cache-ia32.cc (revision 10860) |
+++ src/ia32/stub-cache-ia32.cc (working copy) |
@@ -44,27 +44,45 @@ |
Code::Flags flags, |
StubCache::Table table, |
Register name, |
+ Register receiver, |
Register offset, |
Register extra) { |
ExternalReference key_offset(isolate->stub_cache()->key_reference(table)); |
ExternalReference value_offset(isolate->stub_cache()->value_reference(table)); |
+ ExternalReference map_offset(isolate->stub_cache()->map_reference(table)); |
Label miss; |
+ // Multiply by 3. |
+ __ lea(offset, Operand(offset, offset, times_2, 0)); |
+ |
if (extra.is_valid()) { |
// Get the code entry from the cache. |
- __ mov(extra, Operand::StaticArray(offset, times_2, value_offset)); |
+ __ mov(extra, Operand::StaticArray(offset, times_1, value_offset)); |
// Check that the key in the entry matches the name. |
- __ cmp(name, Operand::StaticArray(offset, times_2, key_offset)); |
+ __ cmp(name, Operand::StaticArray(offset, times_1, key_offset)); |
__ j(not_equal, &miss); |
+ // Check the map matches. |
+ __ mov(offset, Operand::StaticArray(offset, times_1, map_offset)); |
+ __ cmp(offset, FieldOperand(receiver, HeapObject::kMapOffset)); |
+ __ j(not_equal, &miss); |
+ |
// Check that the flags match what we're looking for. |
__ mov(offset, FieldOperand(extra, Code::kFlagsOffset)); |
__ and_(offset, ~Code::kFlagsNotUsedInLookup); |
__ cmp(offset, flags); |
__ j(not_equal, &miss); |
+#ifdef DEBUG |
+ if (FLAG_test_secondary_stub_cache && table == StubCache::kPrimary) { |
+ __ jmp(&miss); |
+ } else if (FLAG_test_primary_stub_cache && table == StubCache::kSecondary) { |
+ __ jmp(&miss); |
+ } |
+#endif |
+ |
// Jump to the first instruction in the code stub. |
__ add(extra, Immediate(Code::kHeaderSize - kHeapObjectTag)); |
__ jmp(extra); |
@@ -75,11 +93,19 @@ |
__ push(offset); |
// Check that the key in the entry matches the name. |
- __ cmp(name, Operand::StaticArray(offset, times_2, key_offset)); |
+ __ cmp(name, Operand::StaticArray(offset, times_1, key_offset)); |
__ j(not_equal, &miss); |
+ // Check the map matches. |
+ __ mov(offset, Operand::StaticArray(offset, times_1, map_offset)); |
+ __ cmp(offset, FieldOperand(receiver, HeapObject::kMapOffset)); |
+ __ j(not_equal, &miss); |
+ |
+ // Restore offset register. |
+ __ mov(offset, Operand(esp, 0)); |
+ |
// Get the code entry from the cache. |
- __ mov(offset, Operand::StaticArray(offset, times_2, value_offset)); |
+ __ mov(offset, Operand::StaticArray(offset, times_1, value_offset)); |
// Check that the flags match what we're looking for. |
__ mov(offset, FieldOperand(offset, Code::kFlagsOffset)); |
@@ -87,9 +113,17 @@ |
__ cmp(offset, flags); |
__ j(not_equal, &miss); |
+#ifdef DEBUG |
+ if (FLAG_test_secondary_stub_cache && table == StubCache::kPrimary) { |
+ __ jmp(&miss); |
+ } else if (FLAG_test_primary_stub_cache && table == StubCache::kSecondary) { |
+ __ jmp(&miss); |
+ } |
+#endif |
+ |
// Restore offset and re-load code entry from cache. |
__ pop(offset); |
- __ mov(offset, Operand::StaticArray(offset, times_2, value_offset)); |
+ __ mov(offset, Operand::StaticArray(offset, times_1, value_offset)); |
// Jump to the first instruction in the code stub. |
__ add(offset, Immediate(Code::kHeaderSize - kHeapObjectTag)); |
@@ -159,12 +193,13 @@ |
Register name, |
Register scratch, |
Register extra, |
- Register extra2) { |
+ Register extra2, |
+ Register extra3) { |
Label miss; |
- // Assert that code is valid. The shifting code relies on the entry size |
- // being 8. |
- ASSERT(sizeof(Entry) == 8); |
+ // Assert that code is valid. The multiplying code relies on the entry size |
+ // being 12. |
+ ASSERT(sizeof(Entry) == 12); |
// Assert the flags do not name a specific type. |
ASSERT(Code::ExtractTypeFromFlags(flags) == 0); |
@@ -176,10 +211,14 @@ |
ASSERT(!extra.is(name)); |
ASSERT(!extra.is(scratch)); |
- // Assert scratch and extra registers are valid, and extra2 is unused. |
+ // Assert scratch and extra registers are valid, and extra2/3 are unused. |
ASSERT(!scratch.is(no_reg)); |
ASSERT(extra2.is(no_reg)); |
+ ASSERT(extra3.is(no_reg)); |
+ Counters* counters = masm->isolate()->counters(); |
+ __ IncrementCounter(counters->megamorphic_stub_cache_probes(), 1); |
+ |
// Check that the receiver isn't a smi. |
__ JumpIfSmi(receiver, &miss); |
@@ -190,7 +229,7 @@ |
__ and_(scratch, (kPrimaryTableSize - 1) << kHeapObjectTagSize); |
// Probe the primary table. |
- ProbeTable(isolate(), masm, flags, kPrimary, name, scratch, extra); |
+ ProbeTable(isolate(), masm, flags, kPrimary, name, receiver, scratch, extra); |
// Primary miss: Compute hash for secondary probe. |
__ mov(scratch, FieldOperand(name, String::kHashFieldOffset)); |
@@ -202,11 +241,13 @@ |
__ and_(scratch, (kSecondaryTableSize - 1) << kHeapObjectTagSize); |
// Probe the secondary table. |
- ProbeTable(isolate(), masm, flags, kSecondary, name, scratch, extra); |
+ ProbeTable( |
+ isolate(), masm, flags, kSecondary, name, receiver, scratch, extra); |
// Cache miss: Fall-through and let caller handle the miss by |
// entering the runtime system. |
__ bind(&miss); |
+ __ IncrementCounter(counters->megamorphic_stub_cache_misses(), 1); |
} |